"""
Workers that run in a separate thread.
"""
from ciocore import conductor_submit, exceptions
import sys
import time
import logging
import copy
import traceback
import json_stream
import random

from PySide2.QtCore import QRunnable, QObject, Signal


logger = logging.getLogger(__name__)


class WorkerSignals(QObject):
    on_start = Signal(dict)
    on_job_start = Signal(dict)
    on_progress = Signal(dict)
    on_response = Signal(dict)
    on_error = Signal(dict)
    on_done = Signal()


class SubmissionWorkerBase(QRunnable):
    @classmethod
    def create(
        cls,
        submissions,
        job_count,
        use_mock=False,
        mock_file="",
        mock_frequency=0.5,
    ):
        """Factory method to create a SubmissionWorker or MockSubmissionWorker."""
        if use_mock:
            return MockSubmissionWorker(mock_file, mock_frequency)
        return SubmissionWorker(submissions, job_count)

    def __init__(self):
        """
        Set up common submission worker attributes.
        """
        super().__init__()
        self.signals = WorkerSignals()

        self._is_cancelled = False
        self._raise_exception = False

        self.current_submit_index = 0
        self.current_submit_title = ""

    def progress_handler(self, progress):
        """
        Emit the progress signal.

        Embellish dict with the current job number and title.
        """
        result = self._progress_dict(progress)

        result.update(
            {
                "job_index": self.current_submit_index,
                "job_title": self.current_submit_title,
            }
        )

        self.signals.on_progress.emit(result)

    def error_handler(self, exception_type, exception, tb):
        """
        Emit the error signal.

        Embellish dict with the current job number and title.
        """
        traceback_string = "".join(traceback.format_tb(tb))
        exception_type.__name__

        error = {
            "body": f"job submission failed.",
            "exception": exception,
            "traceback": traceback_string,
            "exception_type": exception_type.__name__,
            "job_title": self.current_submit_title,
            "status": "error",
            "response_code": 500,
        }

        self.signals.on_error.emit(error)

    def response_handler(self, response, response_code):
        response.update(
            {
                "response_code": response_code,
                "job_title": self.current_submit_title,
            }
        )

        self.signals.on_response.emit(response)

    def cancel(self):
        self._is_cancelled = True

    def raise_exception(self):
        self._raise_exception = True

    def emit_start(self, job_count):
        """Emit the start signal."""
        self.signals.on_start.emit({"job_count": job_count})

    def emit_job_start(self):
        """Emit the job start signal."""
        self.signals.on_job_start.emit(
            {
                "job_index": self.current_submit_index,
                "job_title": self.current_submit_title,
            }
        )

    def emit_done(self):
        """Emit the done signal."""
        self.signals.on_done.emit()

    @staticmethod
    def _progress_dict(progress):
        """Convert upload_stats object to a plain old (JSONifyable) dict."""
        if not hasattr(progress, "files_to_analyze"):
            return {}
        dikt = {
            "files_to_analyze": progress.files_to_analyze or 0,
            "files_to_upload": progress.files_to_upload or 0,
            "bytes_to_upload": progress.bytes_to_upload.value,
            "bytes_uploaded": progress.bytes_uploaded.value,
            "transfer_rate": progress.transfer_rate.value or 0,
            "file_progress": copy.deepcopy(progress.file_progress.value),
        }

        if progress.percent_complete.value:
            dikt["percent_complete"] = progress.percent_complete.value * 100.0
        else:
            dikt["percent_complete"] = 0

        if progress.time_remaining and progress.time_remaining.value:
            dikt["time_remaining"] = progress.time_remaining.value.total_seconds()
        if progress.elapsed_time:
            dikt["elapsed_time"] = progress.elapsed_time.total_seconds()

        if dikt["files_to_analyze"] > 0:
            dikt["processed_md5s"] = len(
                [f["md5"] for f in dikt["file_progress"].values() if f["md5"]]
            )
            dikt["md5_percent"] = (dikt["processed_md5s"] * 100.0) / dikt[
                "files_to_analyze"
            ]
        else:
            dikt["md5_percent"] = 100
        return dikt


class SubmissionWorker(SubmissionWorkerBase):
    """The worker that submits jobs to Conductor."""

    def __init__(self, submissions, job_count):
        super(SubmissionWorker, self).__init__()
        self.current_submit_function = None
        self.submissions = submissions
        self.job_count = job_count

    def run(self):
        i = 0
        self.emit_start(self.job_count)
        for submission in self.submissions:
            self.current_submit_index = i
            self.current_submit_title = submission["job_title"]
            # print("Submission index: ", self.current_submit_index)
            # print("Submission title: ", self.current_submit_title)
            self.emit_job_start()

            try:
                self.current_submit_function = conductor_submit.Submit(submission)
                self.current_submit_function.progress_handler = self.progress_handler
                response, responseCode = self.current_submit_function.main()
            except exceptions.UserCanceledError:
                self.error_handler(*sys.exc_info())
                break
            except Exception:
                self.error_handler(*sys.exc_info())
                continue

            self.response_handler(response, responseCode)

            i += 1

        self.emit_done()

    def cancel(self):
        super(SubmissionWorker, self).cancel()
        if self.current_submit_function:
            self.current_submit_function.stop_work()


class MockSubmissionWorker(SubmissionWorkerBase):
    """
    Mock submission worker simulates a submission worker.

    It reads the entire submission from a json file and emits the events it contains at a given frequency.

    It is activated by the use_mock flag in the advanced section in dev mode.
    If use_mock is True, then the user has the ability to simulate an error by checking the raise exception box.

    """

    def __init__(self, mock_file, mock_frequency):
        super(MockSubmissionWorker, self).__init__()
        self.mock_file = mock_file
        self.mock_frequency = mock_frequency

    def run(self):
        with open(self.mock_file, "r") as infile:
            data = json_stream.load(infile)
            abort_job = False
            for event in data.persistent():
                converted_event = json_stream.to_standard_types(event)
                evt_type = converted_event["event"]
                if abort_job and evt_type != "job_start":
                    continue
                abort_job = False
                evt_data = converted_event.get("data")
                if evt_type == "start":
                    self.signals.on_start.emit(evt_data)
                elif evt_type == "job_start":
                    self.current_submit_index = evt_data["job_index"]
                    self.current_submit_title = evt_data["job_title"]
                    self.signals.on_job_start.emit(evt_data)
                elif evt_type == "progress":
                    self.signals.on_progress.emit(evt_data)
                elif evt_type == "response":
                    self.signals.on_response.emit(evt_data)
                elif evt_type == "error":
                    self.signals.on_error.emit(evt_data)
                elif evt_type == "done":
                    self.signals.on_done.emit()

                try:
                    if self._is_cancelled:
                        raise exceptions.UserCanceledError(
                            "Job submission was cancelled by the user"
                        )
                    if self._raise_exception:
                        self._raise_exception = False
                        self._raise_random_exception()
                except exceptions.UserCanceledError:
                    self.error_handler(*sys.exc_info())
                    self.signals.on_done.emit()
                    break
                except Exception:
                    self.error_handler(*sys.exc_info())
                    abort_job = True

                time.sleep(self.mock_frequency)

    @staticmethod
    def _raise_random_exception():
        exception_types = [
            ValueError,
            TypeError,
            IndexError,
            KeyError,
            FileNotFoundError,
        ]
        chosen_exception = random.choice(exception_types)
        raise chosen_exception("This is a random exception!")

    @staticmethod
    def _progress_dict(progress):
        return progress
